#include<assert.h>
//! \file memory4.cpp
//   - .      .
// (c) Proteus       lawnmower-man@mail.ru


#define PAGE_SIZE 4096	//!<  
#define PAGE_USE   (PAGE_SIZE-sizeof(void*)*4-sizeof(I)*2)	//!<     
#define PAGE_COUNT ((PAGE_USE)/sizeof(item_st))				//!<    


#define LIST_LINK 0		//!<   
#define FREE_LINK 1		//!<   

typedef unsigned char byte;
typedef unsigned short word;

#pragma pack(push,1)

//!  
//! \param T  
//! \param I    (  )
template<typename T,typename I=word>class Allocator{
	//!  
	typedef struct{
		union{
			T obj;			//!<   
			I next_free;	//!<    ,   
		};
		byte id;	//!< c       
	}item_st;
	//!  
	struct bank_st{
		union {
			item_st array[1];	//!<    (    )
			struct {
				char undef[PAGE_USE];	//!<      
				I free_index;			//!<      
				I free_count;			//!<   
				bank_st *prev[2];		//!<    (  )
				bank_st *next[2];		//!<    (  )
			};
		};
	};
	bank_st *bank_list;										//!<   
	bank_st *free_list;										//!<   
	void NewBank();											//!<    
	bank_st *ListInsert(bank_st*,bank_st*,unsigned int i);	//!<      
	bank_st *ListRemove(bank_st*,unsigned int i);			//!<     
public:
	Allocator();
	~Allocator();
	T *New();					//!<  
	void Free(T*);				//!<  
	void Clear();				//!<     
	inline bool isCorrect();	//!<    (   )
	void Stat();				//!<   
};
#pragma pack(pop)


template<class T,class I>Allocator<T,I>::Allocator()
{
	assert(isCorrect());
	bank_list=0;
	free_list=0;
}


template<class T,class I>Allocator<T,I>::~Allocator()
{
	Clear();
}

//!  .       
template<class T,class I>void Allocator<T,I>::Clear()
{
	bank_st *next;
	free_list=0;
	if(bank_list) bank_list->prev[LIST_LINK]->next[LIST_LINK]=0;
	while(bank_list) {
		next=bank_list->next[LIST_LINK];
		delete bank_list;
		bank_list=next;
	}
}

//!     
//! \param list 
//! \param bank  
//! \param i        (   )
//! \result    (      )
template<class T,class I>typename Allocator<T,I>::bank_st *Allocator<T,I>::ListInsert(bank_st *list,bank_st *bank,unsigned int i)
{
	if(!list) list=bank->next[i]=bank->prev[i]=bank;
	else {
		bank->prev[i]=list->prev[i];
		bank->next[i]=list;
		list->prev[i]=bank;
		bank->prev[i]->next[i]=bank;
	}
	return list;
}

//!     
//! \param list  (  )
//! \param i        (   )
//! \result   
template<class T,class I>typename Allocator<T,I>::bank_st *Allocator<T,I>::ListRemove(bank_st *list,unsigned i)
{
	bank_st *prev;
	if(!list || list->next[i]==list) return 0;
	prev=list->prev[i];
	list=list->next[i];
	prev->next[i]=list;
	list->prev[i]=prev;
	return list;
}

//!      
template<class T,class I>void Allocator<T,I>::NewBank()
{
	unsigned int n;
	//  
	bank_st *bank=new bank_st;
	//     (     )
	bank_list=ListInsert(bank_list,bank,LIST_LINK);
	free_list=ListInsert(free_list,bank,FREE_LINK);
	free_list->free_index=0;
	free_list->free_count=(I)PAGE_COUNT;
	//    
	for(n=0; n<PAGE_COUNT-1; n++) free_list->array[n].next_free=n+1;
	free_list->array[n].next_free=-1;
	for(n=0; n<PAGE_COUNT; n++) free_list->array[n].id=(byte)(n>(byte)-1?-1:n);
}

//!   (     )
template<class T,class I>void Allocator<T,I>::Free(T *obj)
{
	item_st *item;
	//   
	unsigned int index=0;
	item=(item_st*)obj;
	while(item->id) index+=item->id,item-=item->id;
	bank_st *bank=(bank_st*)(item);
	//     ,      
	item=(item_st*)obj;
	if(bank->free_index==(I)-1) free_list=ListInsert(free_list,bank,FREE_LINK);
	//      
	item->next_free=bank->free_index;
	bank->free_index=index;
	bank->free_count++;
	//      ,   )
	if(bank->free_count==PAGE_COUNT) {
		bank_list=ListRemove(bank,LIST_LINK);
		bank_st *temp=ListRemove(bank,FREE_LINK);
		if(bank==free_list) free_list=temp; //     if(!temp) free_list=0,     
		delete bank;
	}
	//            (     )
	else if(bank->free_count<free_list->free_count) free_list=bank;
}


//!   
template<class T,class I>T* Allocator<T,I>::New()
{
	//    ,  
	if(!free_list) NewBank();
	//     
	item_st *item=&free_list->array[free_list->free_index];
	free_list->free_index=item->next_free;
	free_list->free_count--;
	//      ,      
	if(free_list->free_index==(I)-1) free_list=ListRemove(free_list,FREE_LINK);
	// ..
	return &item->obj;
}

//!    (     )
//! \result false    
template<class T,class I>inline bool Allocator<T,I>::isCorrect()
{
	return (I)-1>PAGE_COUNT && sizeof(T)<PAGE_SIZE/3;
}


template<class T,class I>void Allocator<T,I>::Stat()
{
#ifdef _DEBUG
	bank_st *bank;
	int count,lost,all_items=0,free_items=0;

	count=0; bank=bank_list;
	if(bank) do {
		count++;
		all_items+=PAGE_COUNT;
		bank=bank->next[LIST_LINK];
	}while(bank!=bank_list);
	printf("pages count     : %d\n",count);		//  
	
	count=0; bank=free_list;
	if(bank) do {
		count++;
		free_items+=bank->free_count;
		bank=bank->next[FREE_LINK];
	}while(bank!=free_list);
	lost=PAGE_USE-(PAGE_COUNT*sizeof(item_st));
	printf("free pages count: %d\n",count);			//    
	printf("all items count : %d\n",all_items);		//    
	printf("free items count: %d\n",free_items);	//     
	printf("bank lost bytes : %d\n",lost);			//       
#endif
}


#include<stdio.h>
#include<memory.h>
#include<stdlib.h>
#define SIZE 5000


int main()
{
// 
	int n,i;
	int *test[SIZE]; memset(test,0,sizeof(test));
	Allocator<int> obj;
	for(n=0; n<SIZE*1000; n++) {
		i=rand()%SIZE;
		printf("pass: %d(%02d%%)\r",n,(n+1)*100/(SIZE*1000));
		if(test[i]) {
			if(*test[i]==i) {printf("Internal error\n\a");return 1;}
			//printf("free: %d\n",i);
			obj.Free(test[i]),test[i]=0;
		}else{
			//printf("make: %d\n",i);
			test[i]=obj.New();
		}
	}
	putchar('\n');
	obj.Stat();
	getchar();
	return 0;
}
